function mark_targets(targets, maelstroms, maelstrom_data, cycle)
{
	if (targets.length > 0) for (var scheme in schemes)
	{
		// Make shallow copies so that schemes are free to sort or delete
		// Command objects tracks totals for convenience
		var ts = targets.slice();
		
		for (var mname in maelstroms)
		{
			// Reset the singular used-flags (keeping the named marks)
			ts.forEach((t) => t.a.used = t.b.used = false);
			
			// Run the scheme upon this maelstrom
			schemes[scheme](maelstroms[mname].slice(), new ThargCmd(ts, scheme, cycle, maelstrom_data[mname]));
			
		}
	}
}

async function load_data()
{
	const el = document.getElementById('main');
	
	const mael_titles = document.createElement('div');
	mael_titles.className = 'mael-titles';
	el.append(mael_titles);
	
	for (var i = cycleFirst; i <= cycleLast; i += 1)
	{
		const cycle = i;
		
		await fetch('data/cycle' + cycle + '.json?2')
		.then((res) => res.json())
		.then((obj) =>
		{
			var cyc = document.createElement('div');
			cyc.id = 'cycle-'+cycle;
			cyc.className = 'cycle';
			
			var cyc_title = document.createElement('div');
			cyc_title.className = 'cycle-title';
			cyc_title.innerHTML = 'Cycle '+cycle;
			cyc.append(cyc_title);
			
			// Attack highlights
			for (var atk of obj.attacks)
			{
				var old = document.getElementById('row-'+(cycle-1)+'-'+atk.b.addr);
				if (old) old.classList.add('alert');
			}
			
			
			// Maelstrom attributes
			var maelstrom_data = {};
			for (var mael of obj.maelstroms)
			{
				// name, controls, radius
				maelstrom_data[mael.name] = mael;
			}
			
			
			
			// Split into maelstrom sections
			// Also generate target mark array
			var maelstrom_name = null;
			var maelstrom_attacks = [];
			var maelstroms = {};
			var temp_a = {};
			var temp_b = {};
			for (var tgt of obj.targets)
			{
				if (tgt.a.dist <= 0.0) continue;
				
				if (maelstrom_name == null) maelstrom_name = tgt.m.name;
				else if (maelstrom_name != tgt.m.name)
				{
					maelstroms[maelstrom_name] = maelstrom_attacks;
					maelstrom_attacks = [];
					maelstrom_name = tgt.m.name;
					temp_a = {};
					temp_b = {};
				}
				
				// Unify systems only to have one instance each
				// Also inject "system used" flags (for both attackers and targets)
				var key_a = ''+tgt.a.addr;
				var key_b = ''+tgt.b.addr;
				if (temp_a[key_a]) tgt.a = temp_a[key_a]; else {tgt.a.used = false; tgt.a.targets = []; tgt.a.mark = []; temp_a[key_a] = tgt.a;}
				if (temp_b[key_b]) tgt.b = temp_b[key_b]; else {tgt.b.used = false; tgt.b.targets = []; tgt.b.mark = []; temp_b[key_b] = tgt.b;}
				
				tgt.a.targets.push(tgt);
				tgt.b.targets.push(tgt);
				maelstrom_attacks.push(tgt);
			}
			maelstroms[maelstrom_name] = maelstrom_attacks;
			
			
			if (i == cycleFirst)
			{
				var r = 2;
				for (var mname in maelstroms)
				{
					var mael_title = document.createElement('div');
					mael_title.className = 'mael-title';
					mael_title.id = 'mael-title-'+mname;
					mael_title.innerHTML = mname;
					mael_title.style.gridRow = r++;
					mael_titles.append(mael_title);
				}
			}
			
			mark_targets(obj.targets, maelstroms, maelstrom_data, cycle);
			
			
			for (var mname in maelstroms)
			{
				var m = maelstroms[mname];
				var mael = document.createElement('div');
				mael.id = 'mael-'+mname;
				mael.className = 'mael';
				
				var b_addr = null;
				var m_count = 0;
				
				
				for (var tgt of m)
				{
					if (tgt.evict || !tgt.elig) continue;
					if (tgt.b.addr == b_addr) continue;
					b_addr = tgt.b.addr;
					
					m_count += 1;
					if (m_count > (displayMaxOverride[tgt.m.name] ? displayMaxOverride[tgt.m.name] : displayMax)) break;
					
					
					var row = document.createElement('div');
					row.id = 'row-'+cycle+'-'+tgt.b.addr;
					row.className = 'row';
					
					var sys = document.createElement('div');
					sys.className = 'sys'+(tgt.b.pop ? ' inhab' : '');
					sys.innerHTML = tgt.b.name;
					row.append(sys);
					
					for (var scheme in schemes)
					{
						var mlbl = document.createElement('div');
						var marked = tgt.b.mark.find((s) => s.s == scheme);
						if (marked)
						{
							mlbl.title = '' + scheme;
							mlbl.innerHTML = ''+marked.n;
						}
						mlbl.className = 'mark '+(marked ? 'set' : 'unset');
						row.append(mlbl);
					}
					
					
					mael.append(row);
				}
				cyc.append(mael);
			}
			el.append(cyc);
			
			
		});
		
	}
}

window.onload = load_data;
