
// Helpers
// Standard rules for a target entry being viable
function target_ok(tgt) {return !tgt.evict  && tgt.elig && !tgt.a.used && !tgt.b.used;}
function d2(sys1, sys2)
{
	const dx = sys1.pos.x-sys2.pos.x;
	const dy = sys1.pos.y-sys2.pos.y;
	const dz = sys1.pos.z-sys2.pos.z;
	return dx*dx + dy*dy + dz*dz;
}

// Wrappers
// Replacing the function for one maelstrom only
function one_mael(maelName, oneScheme, otherScheme)
{
	return (targets, cmd) =>
	{
		(cmd.maelstrom() == maelName ? oneScheme : otherScheme)(targets, cmd);
	};
}


// Schemes

// Standard targeting
function standard(targets, cmd)
{
	for (var tgt of targets) if (target_ok(tgt)) if (cmd.attack(tgt)) break;
}


// NAF targeting
function naf_sort(list, inhab_first)
{
	list.sort((tgt_i, tgt_j) =>
	{
		// For different attackers, put near attacker first
		if (tgt_i.a.addr!=tgt_j.a.addr) return tgt_i.a.dist-tgt_j.a.dist;
		// For same attacker, put inhab target ahead of unpop
		else if (inhab_first && (tgt_i.b.pop != tgt_j.b.pop)) return tgt_i.b.pop ? -1 : 1;
		// For same target pop, put far target first
		else return tgt_j.b.dist-tgt_i.b.dist;
	});
}

// Pure NAF
function naf1(targets, cmd)
{
	// For pure NAF just do the NAF sort then use standard targeting
	naf_sort(targets, false);
	standard(targets, cmd);
}

// Inhab first NAF
function naf2(targets, cmd)
{
	naf_sort(targets, true);
	standard(targets, cmd);
}

// NAF with crowd trigger
function crowd_naf(targets, cmd)
{
	// Begin normally, checking for crowds
	var alts = null;
	for (var i=0; i<targets.length; i+=1)
	{
		var tgt = targets[i];
		if (!target_ok(tgt)) continue;
		
		// Find other targets for this attacker
		// Also store them all - will be needed later if any exist
		alts = [];
		for (var j=i+1; j<targets.length; j+=1)
		{
			var alt = targets[j];
			if (target_ok(alt) && alt.a.addr == tgt.a.addr && alt.b.pop) alts.push(alt);
		}
		
		if (alts.length > 0)
		{
			// Multiple attackers, defer and go to NAF stage
			// Also placing the deferred attack into the alt set should cover both rules
			cmd.log('Crowd NAF: '+tgt.a.name+' (vs '+tgt.b.name+') can see '+alts[0].b.name);
			alts.unshift(tgt);
			break;
		}
		else
		{
			// Only one attacker, do a normal attack
			cmd.log('Crowd NAF: normal attack '+tgt.a.name+' -> '+tgt.b.name);
			if (cmd.attack(tgt)) return;
		}
		
	}
	
	// If a single target was fine then alts will not be null
	if (alts == null) return;
	
	// NAF phase
	naf_sort(targets, false);
	naf: for (var tgt of targets)
	{
		if (!target_ok(tgt)) continue;
		
		cmd.log('Crowd NAF: closest '+tgt.a.name+' vs farthest '+tgt.b.name);
		
		if (alts != null)
		{
			for (var alt of alts)
			{
				if (alt.b.addr == tgt.b.addr)
				{
					cmd.log('Crowd NAF: substitute '+alt.a.name+' -> '+alt.b.name);
					
					// Use deferred attacker
					// Doesn't matter whether this was the original deferred target or some other target
					if (cmd.attack(alt)) return;
					alts = null;
					continue naf;
				}
			}
			
			// Do a NAF attack, but not if it would be the last one
			if (cmd.remaining() > 1) cmd.attack(tgt); else return;
			
		}
		else
		{
			cmd.log('Crowd NAF: NAF attack '+tgt.a.name+' -> '+tgt.b.name);
			
			// Deferred attack is resolved, just continue attacking
			if (cmd.attack(tgt)) return;
		}
		
	}
	
}

// NAF with shell budget
function shell_naf(targets, cmd)
{
	var shell = 4;
	const shell_size = 5.0;
	const shell_inner = 10.0;
	naf_sort(targets, true);
	
	naf: for (var tgt of targets)
	{
		if (!target_ok(tgt)) continue;
		
		// Skip if this attack uses too many shell points
		var shell_cost = (~~(tgt.a.dist/shell_size)!=~~(tgt.b.dist/shell_size) ? 1 : 0) + (tgt.a.dist < shell_inner ? 1 : 0);
		if (shell_cost > shell) continue;
		
		shell -= shell_cost;
		if (cmd.attack(tgt)) return;
	}
	
}


// NAF with NOC trigger
function noc_test_sys(m, a, b)
{
	var  dx1 = a.pos.x - m.pos.x;
	var  dy1 = a.pos.y - m.pos.y;
	var  dz1 = a.pos.z - m.pos.z;
	var ndx2 = m.pos.x - b.pos.x;
	var ndy2 = m.pos.y - b.pos.y;
	var ndz2 = m.pos.z - b.pos.z;
	
	
	// Dims must differ
	for (var n of [1, 2])
	{
		if (                 dx1 == n*ndy2 || dx1 == ndz2) return true;
		if (n*dy1 == ndx2 ||                n*dy1 == ndz2) return true;
		if (  dz1 == ndx2 || dz1 == n*ndy2               ) return true;
	}
	return false;
	
	// Original any-dim check
	//if (dx1 == ndx2 || dx1 == ndy2 || dx1 == ndz2) return true;
	//if (dy1 == ndx2 || dy1 == ndy2 || dy1 == ndz2) return true;
	//if (dz1 == ndx2 || dz1 == ndy2 || dz1 == ndz2) return true;
	
}
function noc_test(t1, t2)
{
	return noc_test_sys(t1.m, t1.b, t2.b);
}
function noc_naf(targets, cmd)
{
	// Get the standard targets first
	// Don't want to attack them yet, just need to imitate the interface and the flags
	const std = [];
	const cmd2 = cmd.copy();
	cmd2.remaining = () => {return 5 - std.length;};
	cmd2.attack = (tgt) => {std.push(tgt); tgt.a.used = tgt.b.used = true; return std.length >= 5;};
	standard(targets, cmd2);
	if (std.length < 1) return;
	
	// Reset the flags
	std.forEach((tgt) => tgt.a.used = tgt.b.used = false);
	
	
	// Check for NOCs
	
	// This last-index definition would make it check up to the last standard target...
	//const last_index = targets.indexOf(std[std.length-1]);
	// ...but in theory this should go all the way to the end if no 5th was found, because that's what the standard scheme does
	const last_index = std.length < 5 ? targets.length-1 : targets.indexOf(std[std.length-1]);
	// Doesn't affect any predictions yet, but it could do
	
	if (last_index < 0) return; // Error...
	var bad_index = last_index+1;
	var buf = [];
	var end = 1;
	
	phase1: for (var i=0; i<std.length; i+=1)
	{
		var tgt = std[i];
		var tgt_index = targets.indexOf(tgt);
		
		for (var j=0; j<=last_index; j+=1)
		{
			var j_tgt = targets[j];
			
			// Allowing non-elig NOCs helps some Oya cases but harms too many others
			// Block them for now
			if (!j_tgt.elig) continue;
			
			if (tgt_index != j && noc_test(tgt, j_tgt))
			{
				cmd.log('NOC at '+tgt.m.name+': '+j_tgt.b.name + ' and ' + tgt.b.name);
				
				// Set the list limit for allowing normal attacks
				bad_index = Math.min(j, tgt_index);
				
				
				// Chances to swap attackers around
				var secondary = false;
				// Record a quick list of possible attacker addresses
				var attackers = [];
				for (var k=i; k<std.length; k+=1) attackers.push(std[k].a.addr);
				
				alt: for (var k=i; k<std.length; k+=1)
				{
					var k_tgt = std[k];
					
					// Possible swap with the NOC match
					if (j_tgt.b.addr != k_tgt.b.addr && j_tgt.a.addr == k_tgt.a.addr && tgt.a.addr != k_tgt.a.addr)
					{
						cmd.log('Swap with '+k_tgt.a.name + ' vs ' + k_tgt.b.name);
						secondary = true;
						if (cmd.attack(k_tgt)) return;
						continue alt;
					}
					
					// Possible swaps between standard matches
					for (var l=i; l<std.length; l+=1)
					{
						if (k == l) continue;
						var l_index = 1 + targets.indexOf(std[l]);
						if (l_index < 1 || l_index >= targets.length) continue;
						var l_tgt = targets[l_index];
						if (l_tgt.b.addr != std[l].b.addr) continue;
						cmd.log('Check '+std[k].a.name+' -> '+l_tgt.a.name+' vs ' + l_tgt.b.name);
						
						if (target_ok(l_tgt) && (l_tgt.a.addr == std[k].a.addr) && attackers.includes(l_tgt.a.addr))
						{
							cmd.log('Alt with '+l_tgt.a.name+' vs ' + l_tgt.b.name);
							secondary = true;
							if (cmd.attack(l_tgt)) return;
							continue alt;
						}
					}
				}
				
				// Limit NAFs if no swap was found, then move to NAF
				if (!secondary) end = 2;
				
				break phase1;
			}
		}
		
	}
	
	// Resolve standard targets if they occur before the first NOC problem
	for (var tgt of std) if (targets.indexOf(tgt) < bad_index) if (cmd.attack(tgt)) return;
	
	// NAF phase
	naf_sort(targets, true);
	for (var tgt of targets)
	{
		if (target_ok(tgt))
		{
			if (cmd.remaining() < end) return;
			cmd.attack(tgt);
		}
	}
	
}

function reinf1(targets, cmd)
{
	//// Establish the radius
	//var control_radius = 0;
	//// Actually need to allow ineligible entries here so we get the full radius
	//for (var tgt of targets) if (!tgt.evict && tgt.a.dist > control_radius) control_radius = tgt.a.dist;
	
	
	// Standard stage
	// Track the attackers used here because they don't count for escapees later
	var std_attacker = [];
	for (var tgt of targets)
	{
		if (!target_ok(tgt)) continue;
		
		// Radius check
		if (tgt.b.dist <= cmd.radius())
		{
			std_attacker.push(tgt.a.addr);
			if (cmd.attack(tgt)) return;
		}
		else break;
		
	}
	
	
	// NAF stage
	var escapees = [];
	cmd.miss();
	naf_sort(targets, true);
	for (var tgt of targets)
	{
		// Allow used attackers so that we log their escapees...
		if (!tgt.elig || tgt.evict || tgt.b.used) continue;
		// ...but not standard attackers
		if (std_attacker.includes(tgt.a.addr)) continue;
		
		if (escapees != null)
		{
			if (tgt.a.used)
			{
				// Attacker used, log one escapee
				if (!tgt.b.used && !escapees.includes(tgt.b.addr)) escapees.push(tgt.b.addr);
			}
			else
			{
				// Attacker available, get a free attack if it was an escapee (then stop recording/using escapees)
				if (escapees.includes(tgt.b.addr))
				{
					escapees = null;
					cmd.unmiss();
					if (cmd.attack(tgt)) return; else continue;
				}
				
				// Pre check the budget because we need to continue at zero to find more escapees
				if (!cmd.nomore()) cmd.attack(tgt);
			}
			
		}
		else if (!tgt.a.used)
		{
			// Pure NAF now
			if (cmd.attack(tgt)) return;
		}
		
	}
}

function reinf2(targets, cmd)
{
	var log_cyc = 32;
	//// Establish the radius
	//var control_radius = 0;
	//// Actually need to allow ineligible entries here so we get the full radius
	//for (var tgt of targets) if (!tgt.evict && tgt.a.dist > control_radius) control_radius = tgt.a.dist;
	
	
	// Standard stage
	// Track the attackers used here because they don't count for escapees later
	var std_attacker = [];
	for (var tgt of targets)
	{
		if (!target_ok(tgt)) continue;
		
		// Radius check
		if (tgt.b.dist <= cmd.radius())
		{
			std_attacker.push(tgt.a.addr);
			if (cmd.attack(tgt)) return;
		}
		else break;
		
	}
	
	
	// NAF stage
	var escapees = [];
	
	// Simplified NAF sort without inhab/unpop
	targets.sort((tgt_i, tgt_j) =>
	{
		// Put near attacker first...
		if (tgt_i.a.addr!=tgt_j.a.addr) return tgt_i.a.dist-tgt_j.a.dist;
		// ...but keep the targets near-far (+ index the last target later)
		else return tgt_i.b.dist-tgt_j.b.dist;
	});
	
	
	// Cover charge for getting the NAF door open
	cmd.miss();
	
	// Snapshot the remaining budget - this limits the index at which we place the main NAF hits
	var rem = cmd.remaining();
	
	
	// Janky iteration to pull the next chunk from the flattened target list
	// i0 and i1 mark the start and end of the chunk
	// By any means we just want the list of entries with different targets for the next attacker
	for (var i0=0, i1=0; i0 < targets.length; i0=i1)
	{
		// Scan to select the chunk, stopping if the attacker changes
		while ((i1+=1) < targets.length)
		{
			var tgt2 = targets[i1];
			if (targets[i0].a.addr != tgt2.a.addr) break;
		}
		var candidates = targets.slice(i0, i1).filter(target_ok);
		if (candidates.length < 1) continue;
		
		// Main attacker tries to mark the last target, but can only go so far based on the budget...
		var tgt = candidates[Math.min(Math.max(0, rem-1), candidates.length-1)];
		
		// While we have secondary targets tracked, admit the possibility of getting a free hit
		if (escapees != null)
		{
			if (escapees.includes(tgt.b.addr))
			{
				// Already tracked = this main attacker gets the free hit
				escapees = null;
				cmd.unmiss();
				if (cmd.attack(tgt)) return;
			}
			else
			{
				if (cmd.more())
				{
					// NAF hit
					cmd.attack(tgt);
					
					// Backtrack and log escapees
					// The hit we just did gets filtered due to the b.used flag
					for (var esc of candidates) if (!esc.b.used && !escapees.includes(esc.b.addr)) escapees.push(esc.b.addr);
				}
				else
				{
					// Even with no budget, let the main hit an escapee
					// UNKNOWN: Any escapee it can see, or only a specific one?
					// UNKNOWN: Starting with near or far?
					for (var esc of candidates) if (escapees.includes(esc.b.addr))
					{
						escapees = null;
						cmd.unmiss();
						if (cmd.attack(esc)) return;
					}
				}
			}
		}
		else
		{
			// Pure NAF now
			if (cmd.cycle() == log_cyc) console.log(cmd.cycle(), 'Pure NAF', tgt.a, tgt.b);
			if (cmd.attack(tgt)) return;
		}
		
		
	}
	
}


// Which schemes to show
const schemes =
{
	//'Standard':standard,
	//'NAF1':naf1,
	//'NAF2':naf2,
	//'Crowd NAF':crowd_naf,
	//'Shell NAF':shell_naf,
	//'Oya crowd NAF':one_mael('Oya', crowd_naf, standard),
	//'NOC NAF':noc_naf,
	'Reinf 1':reinf1,
	'Reinf 2':reinf2,
};
